import pytest

from ferveo_py import (
    encrypt,
    combine_decryption_shares_simple,
    combine_decryption_shares_precomputed,
    decrypt_with_shared_secret,
    Keypair,
    Validator,
    Dkg,
    AggregatedTranscript,
    DkgPublicKey
)


def gen_eth_addr(i: int) -> str:
    return f"0x{i:040x}"


def decryption_share_for_variant(variant, agg_transcript):
    if variant == "simple":
        return agg_transcript.create_decryption_share_simple
    elif variant == "precomputed":
        return agg_transcript.create_decryption_share_precomputed
    else:
        raise ValueError("Unknown variant")


def combine_shares_for_variant(variant, decryption_shares):
    if variant == "simple":
        return combine_decryption_shares_simple(decryption_shares)
    elif variant == "precomputed":
        return combine_decryption_shares_precomputed(decryption_shares)
    else:
        raise ValueError("Unknown variant")


def scenario_for_variant(variant, shares_num, threshold, shares_to_use):
    if variant not in ["simple", "precomputed"]:
        raise ValueError("Unknown variant: " + variant)

    tau = 1
    validator_keypairs = [Keypair.random() for _ in range(0, shares_num)]
    validators = [
        Validator(gen_eth_addr(i), keypair.public_key())
        for i, keypair in enumerate(validator_keypairs)
    ]
    validators.sort(key=lambda v: v.address)

    messages = []
    for sender in validators:
        dkg = Dkg(
            tau=tau,
            shares_num=shares_num,
            security_threshold=threshold,
            validators=validators,
            me=sender,
        )
        messages.append((sender, dkg.generate_transcript()))

    dkg = Dkg(
        tau=tau,
        shares_num=shares_num,
        security_threshold=threshold,
        validators=validators,
        me=validators[0],
    )
    pvss_aggregated = dkg.aggregate_transcripts(messages)
    assert pvss_aggregated.verify(shares_num, messages)

    dkg_pk_bytes = bytes(dkg.public_key)
    dkg_pk = DkgPublicKey.from_bytes(dkg_pk_bytes)

    msg = "abc".encode()
    aad = "my-aad".encode()
    ciphertext = encrypt(msg, aad, dkg_pk)

    decryption_shares = []
    for validator, validator_keypair in zip(validators, validator_keypairs):
        dkg = Dkg(
            tau=tau,
            shares_num=shares_num,
            security_threshold=threshold,
            validators=validators,
            me=validator,
        )
        pvss_aggregated = dkg.aggregate_transcripts(messages)
        assert pvss_aggregated.verify(shares_num, messages)

        decryption_share = decryption_share_for_variant(variant, pvss_aggregated)(
            dkg, ciphertext, aad, validator_keypair
        )
        decryption_shares.append(decryption_share)

    decryption_shares = decryption_shares[:shares_to_use]

    shared_secret = combine_shares_for_variant(variant, decryption_shares)

    if variant == "simple" and len(decryption_shares) < threshold:
        with pytest.raises(ValueError):
            decrypt_with_shared_secret(ciphertext, aad, shared_secret, dkg.public_params)
        return

    if variant == "precomputed" and len(decryption_shares) < shares_num:
        with pytest.raises(ValueError):
            decrypt_with_shared_secret(ciphertext, aad, shared_secret, dkg.public_params)
        return

    plaintext = decrypt_with_shared_secret(ciphertext, aad, shared_secret, dkg.public_params)
    assert bytes(plaintext) == msg


def test_simple_tdec_has_enough_messages():
    scenario_for_variant("simple", shares_num=4, threshold=3, shares_to_use=3)


def test_simple_tdec_doesnt_have_enough_messages():
    scenario_for_variant("simple", shares_num=4, threshold=3, shares_to_use=2)


def test_precomputed_tdec_has_enough_messages():
    scenario_for_variant("precomputed", shares_num=4, threshold=4, shares_to_use=4)


def test_precomputed_tdec_doesnt_have_enough_messages():
    scenario_for_variant("precomputed", shares_num=4, threshold=4, shares_to_use=3)


PARAMS = [
    (1, 'simple'),
    (4, 'simple'),
    (8, 'simple'),
    (32, 'simple'),
    (1, 'precomputed'),
    (4, 'precomputed'),
    (8, 'precomputed'),
    (32, 'precomputed'),

]

TEST_CASES_WITH_THRESHOLD_RANGE = []
for (shares_num, variant) in PARAMS:
    for threshold in range(1, shares_num):
        TEST_CASES_WITH_THRESHOLD_RANGE.append((variant, shares_num, threshold))


# Avoid running this test case as it takes a long time
# @pytest.mark.parametrize("variant, shares_num, threshold", TEST_CASES_WITH_THRESHOLD_RANGE)
# def test_reproduce_nucypher_issue(variant, shares_num, threshold):
#     scenario_for_variant(variant, shares_num, threshold, shares_to_use=threshold)


if __name__ == "__main__":
    pytest.main(["-v", "-k", "test_ferveo"])



from ferveo_py import (
    Keypair,
    Validator,
    Dkg,
    DkgPublicParameters,
    DkgPublicKey
)


def gen_eth_addr(i: int) -> str:
    return f"0x{i:040x}"


tau = 1
security_threshold = 3
shares_num = 4
validator_keypairs = [Keypair.random() for _ in range(shares_num)]
validators = [
    Validator(gen_eth_addr(i), keypair.public_key())
    for i, keypair in enumerate(validator_keypairs)
]
validators.sort(key=lambda v: v.address)


def make_dkg_public_params():
    me = validators[0]
    dkg = Dkg(
        tau=tau,
        shares_num=shares_num,
        security_threshold=security_threshold,
        validators=validators,
        me=me,
    )
    return dkg.public_params


def make_dkg_public_key():
    me = validators[0]
    dkg = Dkg(
        tau=tau,
        shares_num=shares_num,
        security_threshold=security_threshold,
        validators=validators,
        me=me,
    )
    return dkg.public_key


def make_shared_secret():
    # TODO: implement this
    pass


def test_dkg_public_parameters_serialization():
    dkg_public_params = make_dkg_public_params()
    serialized = bytes(dkg_public_params)
    deserialized = DkgPublicParameters.from_bytes(serialized)
    # TODO: Implement comparison
    # assert dkg_public_params == deserialized


# def test_shared_secret_serialization():
#     shared_secret = create_shared_secret_instance()
#     serialized = bytes(shared_secret)
#     deserialized = SharedSecret.from_bytes(serialized)
#     TODO: Implement comparison
#     assert shared_secret == deserialized

def test_keypair_serialization():
    keypair = Keypair.random()
    serialized = bytes(keypair)
    deserialized = Keypair.from_bytes(serialized)
    # TODO: Implement comparison
    # assert keypair == deserialized


def test_dkg_public_key_serialization():
    dkg_pk = make_dkg_public_key()
    serialized = bytes(dkg_pk)
    assert len(serialized) == DkgPublicKey.serialized_size()



from .ferveo_py import (
    encrypt,
    combine_decryption_shares_simple,
    combine_decryption_shares_precomputed,
    decrypt_with_shared_secret,
    Keypair,
    PublicKey,
    Validator,
    Transcript,
    Dkg,
    Ciphertext,
    DecryptionShareSimple,
    DecryptionSharePrecomputed,
    AggregatedTranscript,
    DkgPublicKey,
    DkgPublicParameters,
    SharedSecret,
)



from ferveo_py import (
    encrypt,
    combine_decryption_shares_simple,
    decrypt_with_shared_secret,
    Keypair,
    Validator,
    Dkg,
    AggregatedTranscript,
)


def gen_eth_addr(i: int) -> str:
    return f"0x{i:040x}"


tau = 1
security_threshold = 3
shares_num = 4
validator_keypairs = [Keypair.random() for _ in range(0, shares_num)]
validators = [
    Validator(gen_eth_addr(i), keypair.public_key())
    for i, keypair in enumerate(validator_keypairs)
]

# Validators must be sorted by their public key
validators.sort(key=lambda v: v.address)

# Each validator holds their own DKG instance and generates a transcript every
# validator, including themselves
messages = []
for sender in validators:
    dkg = Dkg(
        tau=tau,
        shares_num=shares_num,
        security_threshold=security_threshold,
        validators=validators,
        me=sender,
    )
    messages.append((sender, dkg.generate_transcript()))

# Now that every validator holds a dkg instance and a transcript for every other validator,
# every validator can aggregate the transcripts
me = validators[0]
dkg = Dkg(
    tau=tau,
    shares_num=shares_num,
    security_threshold=security_threshold,
    validators=validators,
    me=me,
)

# Server can aggregate the transcripts
server_aggregate = dkg.aggregate_transcripts(messages)
assert server_aggregate.verify(shares_num, messages)

# And the client can also aggregate and verify the transcripts
client_aggregate = AggregatedTranscript(messages)
assert client_aggregate.verify(shares_num, messages)

# In the meantime, the client creates a ciphertext and decryption request
msg = "abc".encode()
aad = "my-aad".encode()
ciphertext = encrypt(msg, aad, dkg.public_key)

# The client can serialize/deserialize ciphertext for transport
ciphertext_ser = bytes(ciphertext)

# Having aggregated the transcripts, the validators can now create decryption shares
decryption_shares = []
for validator, validator_keypair in zip(validators, validator_keypairs):
    dkg = Dkg(
        tau=tau,
        shares_num=shares_num,
        security_threshold=security_threshold,
        validators=validators,
        me=validator,
    )

    # We can also obtain the aggregated transcript from the side-channel (deserialize)
    aggregate = AggregatedTranscript(messages)
    assert aggregate.verify(shares_num, messages)

    # The ciphertext is obtained from the client

    # Create a decryption share for the ciphertext
    decryption_share = aggregate.create_decryption_share_simple(
        dkg, ciphertext, aad, validator_keypair
    )
    decryption_shares.append(decryption_share)

# Now, the decryption share can be used to decrypt the ciphertext
# This part is in the client API

shared_secret = combine_decryption_shares_simple(decryption_shares)

# The client should have access to the public parameters of the DKG

plaintext = decrypt_with_shared_secret(ciphertext, aad, shared_secret, dkg.public_params)
assert bytes(plaintext) == msg

print("Success!")



from typing import Sequence


class Keypair:
    @staticmethod
    def random() -> Keypair:
        ...

    @staticmethod
    def from_secure_randomness(data: bytes) -> Keypair:
        ...

    @staticmethod
    def secure_randomness_size(data: bytes) -> int:
        ...

    @staticmethod
    def from_bytes(data: bytes) -> Keypair:
        ...

    def __bytes__(self) -> bytes:
        ...

    def public_key(self) -> PublicKey:
        ...


class PublicKey:
    @staticmethod
    def from_bytes(data: bytes) -> PublicKey:
        ...

    def __bytes__(self) -> bytes:
        ...


class Validator:

    def __init__(self, address: str, public_key: PublicKey):
        ...

    address: str

    public_key: PublicKey


class Transcript:
    @staticmethod
    def from_bytes(data: bytes) -> Transcript:
        ...

    def __bytes__(self) -> bytes:
        ...


class DkgPublicKey:
    @staticmethod
    def from_bytes(data: bytes) -> DkgPublicKey:
        ...

    def __bytes__(self) -> bytes:
        ...


class Dkg:

    def __init__(
            self,
            tau: int,
            shares_num: int,
            security_threshold: int,
            validators: Sequence[Validator],
            me: Validator,
    ):
        ...

    public_key: DkgPublicKey

    public_params: DkgPublicParameters

    def generate_transcript(self) -> Transcript:
        ...

    def aggregate_transcripts(self, messages: Sequence[(Validator, Transcript)]) -> AggregatedTranscript:
        ...


class Ciphertext:
    @staticmethod
    def from_bytes(data: bytes) -> Ciphertext:
        ...

    def __bytes__(self) -> bytes:
        ...


class DecryptionShareSimple:
    @staticmethod
    def from_bytes(data: bytes) -> DecryptionShareSimple:
        ...

    def __bytes__(self) -> bytes:
        ...


class DecryptionSharePrecomputed:
    @staticmethod
    def from_bytes(data: bytes) -> DecryptionSharePrecomputed:
        ...

    def __bytes__(self) -> bytes:
        ...


class DkgPublicParameters:
    @staticmethod
    def from_bytes(data: bytes) -> DkgPublicParameters:
        ...

    def __bytes__(self) -> bytes:
        ...


class AggregatedTranscript:

    def __init__(self, messages: Sequence[(Validator, Transcript)]):
        ...

    def verify(self, shares_num: int, messages: Sequence[(Validator, Transcript)]) -> bool:
        ...

    def create_decryption_share_simple(
            self,
            dkg: Dkg,
            ciphertext: Ciphertext,
            aad: bytes,
            validator_keypair: Keypair
    ) -> DecryptionShareSimple:
        ...

    def create_decryption_share_precomputed(
            self,
            dkg: Dkg,
            ciphertext: Ciphertext,
            aad: bytes,
            validator_keypair: Keypair
    ) -> DecryptionSharePrecomputed:
        ...


    @staticmethod
    def from_bytes(data: bytes) -> AggregatedTranscript:
        ...

    def __bytes__(self) -> bytes:
        ...


class SharedSecret:

    @staticmethod
    def from_bytes(data: bytes) -> SharedSecret:
        ...

    def __bytes__(self) -> bytes:
        ...


def encrypt(message: bytes, add: bytes, dkg_public_key: DkgPublicKey) -> Ciphertext:
    ...


def combine_decryption_shares_simple(
        decryption_shares: Sequence[DecryptionShareSimple],
        dkg_public_params: DkgPublicParameters,
) -> bytes:
    ...


def combine_decryption_shares_precomputed(
        decryption_shares: Sequence[DecryptionSharePrecomputed],
) -> SharedSecret:
    ...


def decrypt_with_shared_secret(
        ciphertext: Ciphertext,
        aad: bytes,
        shared_secret: SharedSecret,
        dkg_params: DkgPublicParameters,
) -> bytes:
    ...
